#!/usr/bin/env python3

"""
** Adds information and makes verifications. **
-----------------------------------------------

This module provides 2 functions. It allows to encapsulate requests
in order to add information about its function.
But it also allows to make verification at the time of the reception.
"""

import collections

KIND2COMPACT = {
    'request': b'\x00',
    'answer': b'\x01',
    'package': b'\x02',
    'result': b'\x03',
}
COMPACT2KIND = {value: key for key, value in KIND2COMPACT.items()}


def format_package(content, kind, signature=None):
    """
    ** Allows you to send information by indicating its nature. **

    Parameters
    ----------
    content
        The content of the message, an object of any kind.
    kind : str
        The kind of content :

        - *request* : To ask a question.
        - *answer* : For the results of a request.
    signature : bytes, optional
        A unique identifier to direct the message to the right place upon reception.
        It can for example be issued from the function ``uuid.uuid4().bytes``.

    Raises
    ------
    ValueError
        If *kind* is not one of the possible choices.

    Returns
    -------
    message : tuple
        The request encapsulated in a context.
    """
    try:
        compact_kind = KIND2COMPACT[kind]
    except KeyError as err:
        raise ValueError(
            f"'kind' can only take the values {tuple(KIND2COMPACT)}, "
            f'but this value is {kind}'
            ) from err
    if signature is None:
        return (content, compact_kind)
    return (content, compact_kind, signature)

def unpack_message(message):
    """
    ** Allows to separate the message from the context. **

    Does not check the formatting of the message.

    Parameters
    ----------
    message : tuple
        The request encapsulated in a context,
        indirectly from the ``format_package`` function.

    Returns
    -------
    content
        The content of the sent message.
    context : namedtuple
        Contains the context of the object.

    Examples
    --------
    >>> from raisin.encapsulation.formatting import format_package, unpack_message
    >>> unpack_message(format_package('question', 'request'))
    ('question', context(kind='request', signature=None))
    >>> unpack_message(format_package('response', 'answer'))
    ('response', context(kind='answer', signature=None))
    >>> unpack_message(format_package('response', 'answer', signature=b'abcdefg'))
    ('response', context(kind='answer', signature=b'abcdefg'))
    >>>
    """
    content, compact_kind, *signature = message
    kind = COMPACT2KIND[compact_kind]
    Context = collections.namedtuple('context', ['kind', 'signature'])
    if signature:
        return content, Context(kind, signature[0])
    return content, Context(kind, None)



#!/usr/bin/env python3

"""
** Establishes small communications. **
---------------------------------------

Allows via ``raisin.communication.abstraction.SelectiveConn``,
to take small requests or short exchanges.
The connection must in most cases be connected to another listening connection.
That is, indirectly connected to a ``raisin.communication.handler.Handler``.
"""

def hello(conn):
    """
    ** Sends a 'hello' and expects a 'hello' in return. **

    Parameters
    ----------
    conn : raisin.communication.abstraction.SelectiveConn
        A connection to a ``raisin.communication.handler.Handler``.

    Raises
    ------
    ValueError
        If the answer is not the one expected.

    Examples
    --------
    >>> import socket
    >>> from raisin.communication.abstraction import SocketConn
    >>> from raisin.communication.handler import Handler
    >>> from raisin.communication.request import hello
    >>>
    >>> soc1, soc2 = socket.socketpair()
    >>> with SocketConn(soc1) as conn, Handler(SocketConn(soc2)) as handler:
    ...     handler.start()
    ...     hello(conn)
    ...
    'hello'
    >>>
    """
    return conn.dialog('hello')

def send_package(conn, package):
    """
    ** Sends an ``raisin.encapsulation.packaging.Package``. **

    Parameters
    ----------
    conn : raisin.communication.abstraction.SelectiveConn
        A connection to a ``raisin.communication.handler.Handler``.
    package : raisin.encapsulation.packaging.Package
        A package that allows to contribute to the execution of a task.

    Examples
    --------
    >>> import socket
    >>> from raisin.communication.abstraction import SocketConn
    >>> from raisin.communication.handler import Handler
    >>> from raisin.encapsulation.packaging import Argument
    >>> from raisin.communication.request import send_package
    >>>
    >>> arg = Argument(0)
    >>> soc1, soc2 = socket.socketpair()
    >>> with SocketConn(soc1) as conn, Handler(SocketConn(soc2)) as handler:
    ...     handler.start()
    ...     send_package(conn, arg)
    ...
    >>>
    """
    conn.send_formatted(package, kind='package')

def send_result(conn, result):
    """
    ** Sends an ``raisin.encapsulation.packaging.Result``. **

    Parameters
    ----------
    conn : raisin.communication.abstraction.SelectiveConn
        A connection to a ``raisin.communication.handler.Handler``.
    result : raisin.encapsulation.packaging.Result
        The result of the task.

    Examples
    --------
    >>> import socket
    >>> from raisin.communication.abstraction import SocketConn
    >>> from raisin.communication.handler import Handler
    >>> from raisin.encapsulation.packaging import Result
    >>> from raisin.communication.request import send_result
    >>>
    >>> res = Result(0)
    >>> soc1, soc2 = socket.socketpair()
    >>> with SocketConn(soc1) as conn, Handler(SocketConn(soc2)) as handler:
    ...     handler.start()
    ...     send_result(conn, res)
    ...
    >>>
    """
    conn.send_formatted(result, kind='result')



#!/usr/bin/env python3

"""
** Allows for extensive testing around serialization. **
--------------------------------------------------------

There is a function capable of generating many types of objects.
For each of these objects they are serialized then deserialized.
"""

import sys

import pytest

from raisin.serialization.constants import ALPHABET, BUFFER_SIZE
from raisin.serialization import dumps, loads, serialize, deserialize
from raisin.encapsulation.packaging import Argument, Func, Result, Task


# @pytest.mark.skipif(os.name == 'posix', reason="do not run on mac os")
# uuid.uuid4() # doctest: +SKIP


LEN = {
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    8,
    9,
    10,
    15,
    16,
    17,
    31,
    32,
    33,
    63,
    64,
    65,
    127,
    128,
    129,
    255,
    256,
    257,
    1023,
    1024,
    1025,
    2047,
    2048,
    2049,
    BUFFER_SIZE - 1,
    BUFFER_SIZE,
    BUFFER_SIZE + 1,
    2 * BUFFER_SIZE - 1,
    2 * BUFFER_SIZE,
    2 * BUFFER_SIZE + 1,
}


def get_bytes():
    """
    ** Makes somme various bytes string. **
    """
    yield b''
    yield from (bytes([i]) for i in range(256))
    yield from (bytes(l) for l in LEN)

def get_float():
    """
    ** Makes some various floats. **
    """
    yield 0.0
    yield 1.0
    yield 0.123
    yield 0.123456
    yield 0.123456789
    yield 123456789.123456789
    yield sys.float_info.max
    yield -sys.float_info.max
    yield sys.float_info.min
    yield -sys.float_info.min

def get_int():
    """
    ** Makes some various intergers. **
    """
    yield 0
    for i in [2 ** j for j in range(1, 16)]:
        yield 2 ** i - 1
        yield 2 ** i
        yield 2 ** i + 1
        yield -(2 ** i - 1)
        yield -(2 ** i)
        yield -(2 ** i + 1)

def get_list():
    """
    ** Makes some various list. **
    """
    for list_len in [0, 1, 2, 4, 100]:
        yield [None for _ in range(list_len)]

def get_str():
    """
    ** Makes some various str. **
    """
    yield ''
    yield 'a'
    yield 'abcdefghi'
    yield ''.join(chr(i) for i in range(256))

def get_simple_objects():
    """
    ** Yield some vaious simple objects. **
    """
    yield None
    yield from get_bytes()
    yield from get_float()
    yield from get_int()
    yield from get_list()
    yield from get_str()


def test_ser_gen():
    """
    ** Tests ``raisin.serialization.deserialize``. **
    """
    for obj in get_simple_objects():
        assert deserialize(serialize(obj)) == obj

def _test_dump(obj_gen):
    """
    ** Test for given objects. **
    """
    for obj in obj_gen:
        ser = dumps(obj, compresslevel=1)
        assert all(c in ALPHABET for c in ser)
        assert loads(ser) == obj

def test_dumps():
    """
    ** Tests dumps. **
    """
    _test_dump(obj for obj in get_simple_objects() if sys.getsizeof(obj) <= 10000)

@pytest.mark.slow
def test_dumps_slow():
    """
    ** Tests dumps. **
    """
    _test_dump(obj for obj in get_simple_objects() if 10000 < sys.getsizeof(obj) < 1000000)

def test_raisin_class():
    """
    ** Tests raisin class. **
    """
    arg = Argument(0)
    func = Func(lambda x: x ** 2)
    task = Task(hash(func), (hash(arg),))
    result = Result(0)

    assert deserialize(serialize(arg)) == arg
    assert deserialize(serialize(func)) == func
    assert deserialize(serialize(task)) == task
    assert deserialize(serialize(result)) == result

def test_composed_list():
    """
    ** Test the serialization of compound lists. **
    """
    objs = (
        [[]],
        [[[]]],
        [[None]],
        [[], []],
        [[None], [None]])
    for obj in objs:
        assert deserialize(serialize(obj)) == obj



#!/usr/bin/env python3

"""
** Makes the communication medium and its use independent. **
-------------------------------------------------------------

Whether it is tcp sockets, files, mailboxes or a drive like dropbox,
we want to be able to use all these media indifferently...
That's the role of this module.
It allows to add an abstraction layer to unify the communication API.
"""

import abc
import socket
import threading
import time
import uuid

from raisin.serialization import deserialize, serialize
from raisin.serialization.iter_tools import anticipate
from raisin.serialization.core import size2tag, tag2size
from raisin.encapsulation.formatting import format_package, unpack_message


__pdoc__ = {
    'AbstractConn.__enter__' : True,
    'AbstractConn.__exit__' : True,
}


class AbstractConn(abc.ABC):
    """
    ** Allows for easier communication. **

    It is the basic abstract class that allows to make information
    carriers of very different nature communicate in the same way.
    """
    @abc.abstractmethod
    def get_conn_names(self):
        """
        ** Retrieves the names of all accessible connections. **

        Must be overwritten.

        Returns
        -------
        conn_names : set
            All the identifiers of the recipients that
            can be contacted via the keyword *dest*.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    @abc.abstractmethod
    def send(self, raw, dest=None):
        """
        ** Sends a data stream to the recipient. **

        Must be overwritten.

        Parameters
        ----------
        raw : iterator
            A generator that yields byte strings. The division is arbitrary.
            On reception, the packets received are not necessarily split in the same way.
            This flow can typically be generated by the ``raisin.serialization.serialize`` function.
        dest : str, default=None
            The address of the recipient(s) of the message.
            By default (None), the message is sent to the main recipient.

        Raises
        ------
        KeyError
            If the address does not exist.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    def send_obj(self, obj, dest=None, **kwargs):
        """
        ** Sends an object to the recipient. **

        Can be overwritten.

        Parameters
        ----------
        obj
            It is the serialisable object that is sent to the recipient(s).
        dest : str, default=None
            Same as ``AbstractConn.send``.
        **kwargs : dict
            These are the parameters passed to the ``raisin.serialization.serialize`` function.
        """
        self.send(serialize(obj, **kwargs), dest=dest)

    @abc.abstractmethod
    def recv(self, dest=None):
        """
        ** Recovers the raw data sent by the sender. **

        Must be overwritten.

        Parameters
        ----------
        dest : str, default=None
            The address of the sender of the message.
            By default (None), the message is read to the main connection.

        Yields
        ------
        bytes
            The packets of bytes emitted by the ``AbstractConn.send``
            method of the sender. Packet decoding may differ, but the order is preserved.

        Raises
        ------
        ConnectionError
            If the connection is broken.
        KeyError
            If the address does not exist.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    def recv_obj(self, **kwargs):
        """
        ** Recover and deserialize the sent object. **

        Can be overwritten.

        Parameters
        ----------
        **kwargs : dict
            These are the parameters passed to the ``raisin.serialization.deserialize`` function.

        Returns
        -------
        obj
            The emitted object.
        """
        return deserialize(self.recv(), **kwargs)

    @abc.abstractmethod
    def close(self):
        """
        ** Definitely close the connection. **

        Must be overwritten.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    def __del__(self):
        """
        ** Help for the garage-collector. **
        """
        self.close()

    def __enter__(self):
        """
        ** Allows to manage the closure through a context manager. **
        """
        return self

    def __exit__(self, exc_type, value, traceback):
        """
        ** Guaranteed that the connection closes in all cases. **
        """
        self.close()


class SelectiveConn(AbstractConn):
    """
    ** Allows you to receive a specific message. **

    In particular, this class allows to retrieve a message without losing the others.
    For example if a thread is waiting for a response, the message that is intended
    for it must not be intercepted by another thread.
    This class allows to manage these conflicts.
    """
    def __init__(self):
        self.lock_send = threading.Lock()
        self.lock_recv = threading.Lock()
        self.lock_unpack = threading.Lock()
        self.res_buff = {} # to each signature, associates the received object

    def send_formatted(self, content, kind, signature=None, **kwargs):
        """
        ** Allows you to send information by indicating its nature. **

        For example, it is necessary to be able to separate questions from answers.

        Parameters
        ----------
        content
            Forwarded to ``raisin.encapsulation.formatting.format_package``.
        kind : str
            Forwarded to ``raisin.encapsulation.formatting.format_package``.
        signature : bytes
            Forwarded to ``raisin.encapsulation.formatting.format_package``.
        **kwargs : dict
            The parameters directly transmitted to the ``AbstractConn.send_obj`` method.

        Notes
        -----
        This method is thread-safe because it uses a lock.
        """
        message = format_package(content, kind, signature)
        with self.lock_send:
            self.send_obj(message, **kwargs)

    def recv_formatted(self, *, kind=None, signature=None, get_context=False, **kwargs):
        """
        ** Receives an object and makes verifications on its content. **

        Parameters
        ----------
        kind : str or set, optional
            Selects only the packages that match this type of packages.
        signature : bytes, optional
            Select only the package with this signature.
        get_context : boolean
            If True, also returns the context, not just the content.
        **kwargs : dict
            The parameters directly transmitted to the ``AbstractConn.recv_obj`` method.

        Returns
        -------
        content
            The content of the package that meets the selection criteria.
        context : namedtuple
            Returns this value only if *get_context* is True.

        Examples
        --------
        >>> import socket
        >>> from raisin.communication.abstraction import SocketConn
        >>> soc1, soc2 = socket.socketpair()
        >>> with SocketConn(soc1) as conn1, SocketConn(soc2) as conn2:
        ...     conn1.send_formatted(0, kind='answer', signature=b'0')
        ...     conn1.send_formatted(1, kind='answer', signature=b'1')
        ...     conn1.send_formatted(2, kind='answer', signature=b'2')
        ...     conn1.send_formatted('a', kind='request')
        ...     conn1.send_formatted('b', kind='request')
        ...     conn2.recv_formatted(signature=b'1')
        ...     conn2.recv_formatted(kind='request')
        ...     conn2.recv_formatted(kind='request', get_context=True)
        ...     conn2.recv_formatted(signature=b'0')
        ...     conn2.recv_formatted(signature=b'2', get_context=True)
        ...
        1
        'a'
        ('b', context(kind='request', signature=None))
        0
        (2, context(kind='answer', signature=b'2'))
        >>>
        """
        if isinstance(kind, str):
            kind = {kind}
        while True:
            with self.lock_unpack:
                for context in self.res_buff:
                    if kind is None or context.kind in kind:
                        if (signature is None
                            or context._asdict().get('signature', None) == signature
                        ):
                            if get_context:
                                return self.res_buff.pop(context), context
                            return self.res_buff.pop(context)
            if self.lock_recv.locked():
                time.sleep(0.5)
                continue
            with self.lock_recv:
                message = self.recv_obj(**kwargs)
                with self.lock_unpack:
                    content, context = unpack_message(message)
                    self.res_buff[context] = content

    def dialog(self, request, **kwargs):
        """
        ** Send a request and wait for the answer. **

        This method can be called from a thread without any risk
        of interference, because everything is set up to make it work.
        This method is blocking: it waits until the result is available.

        Parameters
        ----------
        request
            The content of the request.
        **kwargs : dict
            The parameters directly transmitted to the ``SelectiveConn.send_formatted``
            and ``SelectiveConn.recv_formatted`` method.

        Returns
        -------
        answer
            The content of the result is linked to the issued request.

        Examples
        --------
        >>> import uuid
        >>> uuid.uuid4() # doctest: +SKIP
        4545454
        >>>
        >>> import socket
        >>> from raisin.communication.abstraction import SocketConn
        >>> from raisin.communication.handler import Handler
        >>> soc1, soc2 = socket.socketpair()
        >>> with SocketConn(soc1) as conn, Handler(SocketConn(soc2)) as handler:
        ...     handler.start()
        ...     conn.dialog('hello')
        ...
        'hello'
        >>>
        """
        signature = uuid.uuid4().bytes
        self.send_formatted(request, kind='request', signature=signature, **kwargs)
        return self.recv_formatted(kind='answer', signature=signature, **kwargs)


class SocketConn(SelectiveConn):
    """
    ** Adds an abstraction layer to TCP sockets. **

    Attributes
    ----------
    conn : socket.socket
        A socket capable of reading and sending data.
    """

    def __init__(self, conn):
        SelectiveConn.__init__(self)
        self.conn = conn
        self._recv_pack = b'' # buffer memory

    def get_conn_names(self):
        """
        ** Implementation of ``AbstractConn.get_conn_names``. **
        """
        return {None}

    def send(self, raw, dest=None):
        """
        ** Implementation of ``AbstactConn.send``. **
        """
        if dest is not None:
            raise KeyError('this connection is linked to a single recipient')
        for is_end, pack in anticipate(raw):
            signed_pack = bytes([is_end]) + size2tag(len(pack)) + pack
            self.conn.sendall(signed_pack)

    def recv(self, dest=None):
        """
        ** Implementation of ``AbstraConn.recv``. **
        """
        def gen():
            while True:
                try:
                    yield self.conn.recv(4096)
                except OSError as err:
                    raise ConnectionError('communication seems to be interrupted') from err
        if dest is not None:
            raise KeyError('this connection is linked to a single recipient')
        iter_pack = iter(gen())
        while True:
            # check for len and last pack
            while not self._recv_pack:
                self._recv_pack += next(iter_pack)
            is_end = bool(self._recv_pack[0])
            self._recv_pack = self._recv_pack[1:]
            size, self._recv_pack, _ = tag2size(pack=self._recv_pack, gen=iter_pack)
            while len(self._recv_pack) < size:
                self._recv_pack += next(iter_pack)
            yield self._recv_pack[:size]
            self._recv_pack = self._recv_pack[size:]
            if is_end:
                break

    def close(self):
        """
        ** Implementation of ``AbstractConn.close``. **
        """
        try:
            self.conn.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.conn.close()


class BridgeConn(SelectiveConn):
    """
    ** Allows one connection to reach others. **

    Attributes
    ----------
    main_conn : AbstractConn
        The main connection, the one that is intended
        when the target is not specified.
    other_conn : dict
        To each connection name, associate the connection.
    """

    def __init__(self, main_conn, other_conn):
        SelectiveConn.__init__(self)
        self.main_conn = main_conn
        self.other_conn = other_conn

    def get_conn_names(self):
        """
        ** Implementation of ``AbstractConn.get_conn_names``. **
        """
        return {None} | set(self.other_conn)

    def send(self, raw, dest=None):
        """
        ** Implementation of ``AbstactConn.send``. **
        """
        if dest is None:
            return self.main_conn.send(raw)
        try:
            return self.other_conn[dest].send(raw)
        except KeyError as err:
            raise KeyError(
                f"only the recipients {list(self.other_conn.keys())} exist, not '{dest}'"
            ) from err

    def recv(self, dest=None):
        """
        ** Implementation of ``AbstraConn.recv``. **
        """
        if dest is None:
            yield from self.main_conn.recv()
        elif dest in self.other_conn:
            yield from self.other_conn[dest].recv()
        else:
            raise KeyError(
                f"only the recipients {list(self.other_conn)} exist, not '{dest}'"
            )

    def close(self):
        """
        ** Implementation of ``AbstractConn.close``. **
        """
        self.main_conn.close()
        for conn in self.other_conn.values():
            conn.close()


class ServerConn(SelectiveConn):
    """
    ** Allows you to bridge the gap between clients. **

    The 'server' class inherits from this class.
    Thus, it is possible to make the different clients connected
    to the same server communicate with each other.

    Attributes
    ----------
    clients_conn : dict
        To each client identifier, associates the connection of this client.
    """

    def __init__(self):
        SelectiveConn.__init__(self)
        self.clients_conn = {}

    def _conn_expansion(self, abstract_conn):
        """
        ** Memorizes and expands a connection. **

        The server calls this method when the handler is created.
        This is to extend the *abstract_conn* connection so that
        the handler can have access to other clients.

        Parameters
        ----------
        abstract_conn : AbstractConn
            A simple and minimalist connection that
            allows you to get in touch with a single client.

        Returns
        -------
        extended_abstract_conn : AbstractConn
            A connection that allows of course to continue to communicate with the customers.
            That is, it has a similar behavior to the *abstract_conn* argument,
            but it also allows to address the connections of other clients.
        """
        self.clients_conn['id client'] = abstract_conn
        return BridgeConn(main_conn=abstract_conn, other_conn=self.clients_conn)

    def get_conn_names(self):
        """
        ** Implementation of ``AbstractConn.get_conn_names``. **
        """
        return {None} | set(self.clients_conn)

    def send(self, raw, dest=None):
        """
        ** Implementation of ``AbstactConn.send``. **
        """
        if dest is None:
            raise KeyError(
                'there is no default receiver, you have to specify: '
                f'{list(self.clients_conn)}'
            )
        try:
            return self.clients_conn[dest].send(raw)
        except KeyError as err:
            raise KeyError(
                f"only the recipients {list(self.clients_conn)} exist, not '{dest}'"
            ) from err

    def recv(self, dest=None):
        """
        ** Implementation of ``AbstraConn.recv``. **
        """
        if dest is None:
            raise KeyError(
                'there is no default receiver, you have to specify: '
                f'{list(self.clients_conn.keys())}'
            )
        if dest not in self.clients_conn:
            raise KeyError(
                f"only the recipients {list(self.clients_conn)} exist, not '{dest}'"
            )
        yield from self.clients_conn[dest].recv()

    def close(self):
        """
        ** Implementation of ``AbstractConn.close``. **
        """
        for conn in self.clients_conn.values():
            conn.close()



#!/usr/bin/env python3

"""
** Processes requests. **
-------------------------

Whether it is a client or a server that asks for something, it doesn't change much.
In all cases it is TCP sockets that request and render services.
That's why once the communication is established,
clients and servers use the same function to communicate.
"""

import threading

from raisin.encapsulation.packaging import Argument, Func, Task, Result
from raisin.communication.request import send_result


__pdoc__ = {
    'Handler.__enter__': True,
    'Handler.__exit__': True
}


class Handler(threading.Thread):
    """
    ** Helps a socket to communicate. **

    Attributes
    ----------
    conn : raisin.communication.abstraction.AbstractConn
        The abstract connection that allows communication.
    """

    def __init__(self, conn):
        """
        Parameters
        ----------
        conn : raisin.communication.abstraction.AbstractConn
            An entity able to communicate.
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self.conn = conn

        self._args = {}
        self._func = {}

    def run(self):
        """
        ** Wait for the requests to answer them. **

        This method must be launched asynchronously by invoking the *start* method.
        It listens for the arrival of a request through the 'conn' attribute.
        As soon as a request arrives, it is processed. Once the request is processed,
        this method starts listening for the next request.
        """
        while True:
            try:
                ask, context = self.conn.recv_formatted(
                    kind={'request', 'package'},
                    get_context=True)
            except ConnectionError:
                break

            if ask == 'hello':
                self.conn.send_formatted('hello', kind='answer', signature=context.signature)
            elif isinstance(ask, Argument):
                self._args[ask.__hash__()] = ask
            elif isinstance(ask, Func):
                self._func[ask.__hash__()] = ask
            elif isinstance(ask, Task):
                # TODO : s'assurer qu'il y ai les arguments et la fonction dispo
                func = self._func[ask.func_hash]
                args = [self._args[arg_hash].get_value() for arg_hash in ask.arg_hashes]
                # TODO : calculer le resultat dans un autre processus
                res = Result(func(*args))
                send_result(self.conn, res)

            else:
                raise NotImplementedError(f"impossible to process {ask}, it's not coded")

    def handler_close(self):
        """
        ** Clean up the connection. **
        """
        self.conn.close()

    def __del__(self):
        """
        ** Help for the garage-collector. **
        """
        self.handler_close()

    def __enter__(self):
        """
        ** Allows to manage the closure through a context manager. **
        """
        return self

    def __exit__(self, exc_type, value, traceback):
        """
        ** Guaranteed that the connection closes in all cases. **
        """
        self.handler_close()



#!/usr/bin/env python3

"""
** Allows for extensive testing around communication. **
--------------------------------------------------------
"""

import socket
import threading
import time
import uuid

import pytest

from raisin.communication.abstraction import SocketConn
from raisin.communication.server import Server
from raisin.communication.client import Client
from raisin.communication.request import hello


def test_socket_abstraction_conn():
    """
    ** Tests all fonctions of the abstraction. **
    """
    socket1, socket2 = socket.socketpair()
    abstraction1, abstraction2 = SocketConn(socket1), SocketConn(socket2)

    abstraction1.send((b'mes', b'sage1',))
    abstraction2.send((b'hello',))
    abstraction2.send((b'',))
    abstraction1.send((b'message2',))
    assert b''.join(abstraction2.recv()) == b'message1'
    assert b''.join(abstraction2.recv()) == b'message2'
    assert b''.join(abstraction1.recv()) == b'hello'
    assert b''.join(abstraction1.recv()) == b''

    abstraction1.send_obj(0)
    abstraction1.send_obj(1)
    abstraction2.send_obj(2)
    abstraction2.send_obj(3)
    assert abstraction1.recv_obj() == 2
    assert abstraction1.recv_obj() == 3
    assert abstraction2.recv_obj() == 0
    assert abstraction2.recv_obj() == 1

    abstraction1.close()
    abstraction2.close()

def test_formatted():
    """
    ** Verifies that the threads do not mix up the packages. **
    """
    def send_and_recv(conn1, conn2):
        """ Designed to be executed multiple times from multiple threads. """
        for _ in range(10):
            signature = uuid.uuid4().bytes
            content = uuid.uuid4().bytes
            conn1.send_formatted(content, kind='answer', signature=signature)
            assert conn2.recv_formatted(signature=signature) == content

    socket1, socket2 = socket.socketpair()
    with SocketConn(socket1) as conn1, SocketConn(socket2) as conn2:
        threads = [threading.Thread(target=send_and_recv, args=(conn1, conn2)) for _ in range(16)]
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

@pytest.mark.slow
def test_simple_server():
    """
    ** Tries to launch a simple server. **
    """
    with Server(9999) as server:
        server.start() # the server is listening
        time.sleep(2)
        server.shutdown() # stops the server

@pytest.mark.slow
def test_simple_client():
    """
    ** Tries to lunch a simple client. **
    """
    with Server(9999) as server:
        server.start()
        time.sleep(2)
        with Client(None, 9999) as client:
            client.start()
            time.sleep(2)

@pytest.mark.slow
def test_client_interface():
    """
    ** Performs several tests on the client interface. **
    """
    with Server(9999) as server:
        server.start()
        time.sleep(2)
        with Client(None, 9999) as client:
            hello(client)



#!/usr/bin/env python3

"""
** Allows for extensive testing around serialization. **
--------------------------------------------------------

There is a function capable of generating many types of objects.
For each of these objects they are serialized then deserialized.
"""

import itertools

import pytest

from raisin.serialization.constants import BUFFER_SIZE
from raisin.serialization.iter_tools import (
    anticipate,
    concat_gen,
    deconcat_gen,
    relocate,
    resize,
    to_gen,
)
from raisin.serialization.core import size2tag, tag2size


def test_anticipate():
    """
    ** Tests ``raisin.serialization.iter_tools.anticipate``. **
    """
    assert list(anticipate([''])) == [(True, '')]
    assert list(anticipate(['', ''])) == [(False, ''), (True, '')]
    assert list(anticipate(['', '', ''])) == [(False, ''), (False, ''), (True, '')]
    with pytest.raises(RuntimeError):
        list(anticipate([]))

def test_concat():
    """
    ** Tests ``raisin.serialization.iter_tools.concat_gen``. **
    """
    for gen in ([], [b''], [b'', b''], [b'a'], [b'a', b'b'], [b'', b'b'], [b'a', b'']):
        assert list(deconcat_gen(gen=concat_gen(gen))) == gen
        assert list(deconcat_gen(pack=b''.join(concat_gen(gen)))) == gen
    for gen in ([b'a'], [b'a', b'b']):
        with pytest.raises(ValueError): # without header
            list(deconcat_gen(pack=b''.join(concat_gen(gen))[1:]))
        with pytest.raises(ValueError):
            list(deconcat_gen(pack=b''.join(concat_gen(gen))[:-1]))

def test_relocate():
    """
    ** Tests ``raisin.serialization.iter_tools.relocate``. **
    """
    iterator = iter([])
    iterator_bis = relocate(iterator)
    with pytest.raises(StopIteration):
        next(iterator_bis)
    with pytest.raises(StopIteration):
        next(iterator)
    for gen in (
        [b'a'],
        [b'a', b'b', b'c'],
        itertools.chain([b'a'], (b'b' * BUFFER_SIZE for _ in range(10))),
    ):
        iterator = iter(gen)
        iterator_bis = relocate(iterator)
        with pytest.raises(StopIteration):
            next(iterator)
        assert next(iterator_bis) == b'a'

def test_resize():
    """
    ** Tests ``raisin.serialization.iter_tools.resize``. **
    """
    for nbr in [0, 1, 2, 10, 100, 1000]:
        for pack_len in [0, 1, 2, 10, 100, 1000]:
            for gen_len in [0, 1, 2, 4]:
                for gen_item_len in [0, 1, 100, 500]:
                    pack = bytes(pack_len)
                    gen = (bytes(gen_item_len) for _ in range(gen_len))
                    if nbr <= pack_len + gen_len * gen_item_len:
                        pack_, gen_ = resize(nbr, pack=pack, gen=gen)
                        assert len(pack_) == nbr
                        assert len(pack_ + b''.join(gen_)) == pack_len + gen_len * gen_item_len
                    else:
                        with pytest.raises(RuntimeError):
                            resize(nbr, pack=pack, gen=gen)

def test_to_gen():
    """
    ** Tests ``raisin.serialization.iter_tools.to_gen``. **
    """
    for size in [1, 2, 1000, BUFFER_SIZE]:
        for gen in (
            [],
            [b''],
            [b'a'],
            [b'a' * size],
            [b'a' * (size - 1)],
            [b'a' * (size + 1)],
            [b'', b''],
            [b'a' * ((2 ** l) % (2 * size)) for l in range(100)],
        ):
            l_in = [len(e) for e in gen]
            l_out = [len(e) for e in to_gen(gen=gen, size=size)]
            assert sum(l_in) == sum(l_out)
            assert all(s == size for s in l_out[:-1])
            assert l_out[-1] <= size
            l_out = [len(e) for e in to_gen(pack=b''.join(gen), size=size)]
            assert sum(l_in) == sum(l_out)
            assert all(s == size for s in l_out[:-1])
            assert l_out[-1] <= size

def test_size_tag():
    """
    ** Tests ``raisin.serialization.core.size2tag``. **
    """
    with pytest.raises(StopIteration):
        tag2size(pack=b'')
    for size in [0, 1, 2, 127, 128, 129, 123456789123456789]:
        pack = size2tag(size)
        assert tag2size(pack=pack)[0] == size
        if len(pack) >= 2:
            with pytest.raises(RuntimeError):
                tag2size(pack=pack[:-1])



#!/usr/bin/env python3

"""
** Client TCP. **
-----------------

As soon as you want to connect to a server, you have
to create a client, even if it is for a small communication.
These clients are meant to be created and deleted in large numbers if needed.
"""

import socket
import threading

from raisin.communication.abstraction import SocketConn
from raisin.communication.handler import Handler


__pdoc__ = {
    'Client.__del__': True,
    'Client.__enter__': True,
    'Client.__exit__': True,
    'Client.__repr__': True,
    'Client.__str__': True
}


class BaseClient(threading.Thread, SocketConn):
    """
    ** TCP Client. **

    This client is both able to listen in ipv4 and ipv6.

    Attributes
    ----------
    host : str
        The ip address of the connection in ipv4 or ipv6.
        It can also be a hostname or a domain name.
    port : int
        The communication port.
    tcp_socket : socket.socket
        The tcp socket which allows low level communication.
    """

    def __init__(self, host, port):
        """
        Parameters
        ----------
        ip : str
            The server address.
        port : int
            The server's listening port.

        Raises
        ------
        ConnectionError
            If we can't connect.
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self.host = host
        self.port = port
        self.tcp_socket = BaseClient._init_tcp_socket(self.host, self.port)
        SocketConn.__init__(self, self.tcp_socket)

    @staticmethod
    def _init_tcp_socket(host, port):
        """
        ** Help for the ``BaseServer.__init__``. **

        Paremeters
        ----------
        port : int
            The port to listen on.

        Returns
        -------
        tcp_socket : socket.socket
            The TCP socket ready to communicate.
        """
        tcp_socket = None
        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM):
            family, socktype, proto, _, sockaddr = res
            try:
                tcp_socket = socket.socket(family, socktype, proto)
            except OSError:
                tcp_socket = None
                continue
            try:
                tcp_socket.connect(sockaddr)
            except OSError:
                tcp_socket.close()
                tcp_socket = None
                continue
            break
        if tcp_socket is None:
            raise ConnectionError('could not open socket')
        return tcp_socket

    def run(self):
        """
        ** Puts the client on asynchronous listening mode. **

        Should not be called as is. It is the call of the
        *start* method that executes run.
        """
        Handler(self).run()

    def shutdown(self):
        """
        Tell the ``BaseClient.run`` loop to stop and wait until it does.
        ``BaseClient.shutdown`` must be called while ``BaseClient.run``
        is running in a different thread otherwise it will deadlock.
        """
        self.client_close()
        while self.is_alive():
            continue

    def client_close(self):
        """
        ** Clean up the client. **

        Should not be called if the client is encapsulated
        in a context manager (*with* statement).
        Can be called several times.
        """
        try:
            self.tcp_socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.tcp_socket.close()


class Client(BaseClient):
    """
    ** Enables you to enrich the ``raisin.communication.client.BaseClient``. **
    """

    def __del__(self):
        """
        ** Help for the garbage-collector. **
        """
        try:
            self.client_close()
        except AttributeError:
            pass

    def __enter__(self):
        """
        ** Prepared for easy closing. **

        Allows you to use the *with* statement which allows
        you to set up a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        ** Stop the client. **

        Goes together with ``Client.__enter__``.
        """
        self.shutdown()

    def __repr__(self):
        """
        ** Gives a simple representation of the client. **
        """
        return f'Client({self.host}, {self.port})'

    def __str__(self):
        r"""
        ** Provides a complete representation of the client. **
        """
        return (
            f'TCP Client:\n'
            f'    host={self.host}\n'
            f'    port={self.port}\n'
            f'    tcp_socket={self.tcp_socket}')
